@@ -676,7 +676,7 @@ let () =
        useful to pass -linkpkg when building archives for example
        (.cma and .cmxa); the "linkpkg" flag allows user to request it
        explicitly. *)
-    flag ["ocaml"; "link"; "linkpkg"] & A"-linkpkg";
+    flag ["ocaml"; "link"; "library"; "linkpkg"] & A"-linkpkg";
     pflag ["ocaml"; "link"] "dontlink" ~doc_param
           (fun pkg -> S[A"-dontlink"; A pkg]);
 
@@ -693,13 +693,19 @@ let () =
       ["c"; "compile"];
     ] in
 
-    (* tags package(X), predicate(X) and syntax(X) *)
+    (* tags package(X), predicate(X) and syntax(X), ppopt(X), ppxopt(X) *)
     List.iter begin fun tags ->
       pflag tags "package" ~doc_param (fun pkg -> S [A "-package"; A pkg]);
       pflag tags "predicate" ~doc_param:"archive"
         (fun pkg -> S [A "-predicates"; A pkg]);
-      if List.mem "ocaml" tags then
-        pflag tags "syntax" ~doc_param:"camlp4o" (fun pkg -> S [A "-syntax"; A pkg])
+      if List.mem "ocaml" tags then begin
+        pflag tags "syntax" ~doc_param:"camlp4o"
+          (fun pkg -> S [A "-syntax"; A pkg]);
+        pflag tags "ppopt" ~doc_param:"pparg"
+          (fun arg -> S [A "-ppopt"; A arg]);
+        pflag tags "ppxopt" ~doc_param:"package,arg"
+          (fun arg -> S [A "-ppxopt"; A arg]);
+      end;
     end all_tags
   end else begin
     try
@@ -788,6 +794,8 @@ let () =
     (fun param -> S [A "-inline-max-unroll"; A param]);
   pflag ["ocaml"; "native"; "compile"] "unbox_closures_factor" ~doc_param:"10"
     (fun param -> S [A "-unbox-closures-factor"; A param]);
+  pflag ["ocaml"; "native"; "compile"] "afl_inst_ratio" ~doc_param:"50"
+    (fun param -> S [A "-afl-inst-ratio"; A param]);
   ()
 
 let camlp4_flags camlp4s =
@@ -876,6 +884,11 @@ flag ["ocaml"; "link"; "library"; "custom"; "byte"] (A "-custom");;
 flag ["ocaml"; "link"; "toplevel"; "custom"; "byte"] (A "-custom");;
 flag ["ocaml"; "compile"; "profile"; "native"] (A "-p");;
 
+flag ["ocamlmklib"; "linkall"] (A "-linkall");;
+flag ["ocamlmklib"; "custom"] (A "-custom");;
+flag ["ocamlmklib"; "failsafe"] (A "-failsafe");;
+flag ["ocamlmklib"; "debug"] (A "-g");;
+
 flag ["ocaml"; "compile"; "no_alias_deps";] (A "-no-alias-deps");;
 flag ["ocaml"; "infer_interface"; "no_alias_deps";] (A "-no-alias-deps");;
 
@@ -899,14 +912,20 @@ begin
 end;;
 
 flag ["ocaml"; "native"; "compile"; "no_float_const_prop";] (A "-no-float-const-prop");
-flag ["ocaml"; "compile"; "keep_docs";] (A "-keep-docs");
-flag ["ocaml"; "infer_interface"; "keep_docs";] (A "-keep-docs");
-flag ["ocaml"; "compile"; "keep_locs";] (A "-keep-locs");
-flag ["ocaml"; "absname"; "compile"] (A "-absname");;
-flag ["ocaml"; "absname"; "infer_interface"] (A "-absname");;
+
+["compile"; "link"; "pack"] |> List.iter begin fun phase ->
+  flag ["ocaml"; phase; "keep_docs";] (A "-keep-docs");
+  flag ["ocaml"; phase; "keep_locs";] (A "-keep-locs");
+end;;
+["compile"; "infer_interface"] |> List.iter begin fun phase ->
+  flag ["ocaml"; phase; "absname"] (A "-absname");
+end;;
+
 flag ["ocaml"; "byte"; "compile"; "compat_32";] (A "-compat-32");;
 flag ["ocaml";"compile";"native";"asm"] & S [A "-S"];;
 
+(* AFL instrumentation *)
+flag ["ocaml"; "compile"; "native"; "afl_instrument"] (A "-afl-instrument");;
 
 (* threads, with or without findlib *)
 flag ["ocaml"; "compile"; "thread"] (A "-thread");;
@@ -996,13 +1015,24 @@ pflag ["ocaml"; "doc"; "man"] "man_section"
 ocaml_lib "ocamlbuildlib";;
 ocaml_lib "ocamlbuildlightlib";;
 
-pflag ["c"; "compile"] "cc" (fun param -> S [A "-cc"; A param]);;
-pflag ["c"; "link"] "cc" (fun param -> S [A "-cc"; A param]);;
+begin
+  let ccflag ~lang ~phase ~flag =
+    pflag [lang; phase] flag (fun param -> S [A ("-"^flag); A param])
+  in
+  ["c"; "ocaml"] |> List.iter (fun lang ->
+  ["compile"; "link"] |> List.iter (fun phase ->
+  ["cc"; "ccopt"; "cclib"] |> List.iter (fun flag ->
+    ccflag ~lang ~phase ~flag)))
+end;;
 
-pflag ["c"; "compile"] "ccopt" (fun param -> S [A "-ccopt"; A param]);;
-pflag ["c"; "link"] "ccopt" (fun param -> S [A "-ccopt"; A param]);;
-
-pflag ["c"; "compile"] "cclib" (fun param -> S [A "-cclib"; A param]);;
-pflag ["c"; "link"] "cclib" (fun param -> S [A "-cclib"; A param]);;
+begin
+  let ocamlmklib_pflag ?doc_param flag =
+    pflag ["ocamlmklib"] flag ?doc_param
+      (fun param -> S [A ("-"^flag); A param]) in
+  ocamlmklib_pflag "cclib";
+  ocamlmklib_pflag "ccopt";
+  ocamlmklib_pflag "rpath";
+  ocamlmklib_pflag "ldopt";
+end;;
 
 end in ()
